/*
 *  $Id: pvrusb2-base.c,v 1.1 2005/04/11 01:02:33 isely Exp $
 *
 *  Copyright (C) 2005 Mike Isely <isely@pobox.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "pvrusb2-base.h"
#include "pvrusb2-io.h"
#include "pvrusb2-hdw.h"
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/slab.h>
#include <asm/semaphore.h>

struct pvr2_main *pvr2_main_create(struct usb_device *dev)
{
    struct pvr2_main *mp;

    mp = kmalloc(sizeof(*mp),GFP_KERNEL);
    if (!mp) return mp;
    memset(mp,0,sizeof(*mp));
    mp->hdw = pvr2_hdw_create(dev);
    if (!mp->hdw) goto fail;
    mp->video_stream.stream = pvr2_hdw_get_video_stream(mp->hdw);
    init_MUTEX(&mp->sem);
    return mp;
 fail:
    if (mp->hdw) pvr2_hdw_destroy(mp->hdw);
    kfree(mp);
    return 0;
}


static void pvr2_main_check_removal(struct pvr2_main *mp)
{
    if (mp->mc_first) return;
    if (!mp->disconnect_flag) return;
    pvr2_hdw_destroy(mp->hdw);
    kfree(mp);
}


void pvr2_main_disconnect(struct pvr2_main *mp)
{
    pvr2_hdw_disconnect(mp->hdw);
    pvr2_main_check_removal(mp);
}


void pvr2_channel_init(struct pvr2_channel *cp,struct pvr2_main *mp)
{
    down(&mp->sem);
    cp->mc_head = mp;
    cp->mc_next = 0;
    cp->mc_prev = mp->mc_last;
    if (mp->mc_last) {
        mp->mc_last->mc_next = cp;
    } else {
        mp->mc_first = cp;
    }
    mp->mc_last = cp;
    up(&mp->sem);
}


static void pvr2_channel_disclaim_stream(struct pvr2_channel *cp)
{
    if (!cp->stream) return;
    pvr2_stream_kill(cp->stream->stream);
    cp->stream->user = 0;
    cp->stream = 0;
}


void pvr2_channel_done(struct pvr2_channel *cp)
{
    struct pvr2_main *mp = cp->mc_head;
    down(&mp->sem);
    pvr2_channel_disclaim_stream(cp);
    if (cp->mc_next) {
        cp->mc_next->mc_prev = cp->mc_prev;
    } else {
        mp->mc_last = cp->mc_prev;
    }
    if (cp->mc_prev) {
        cp->mc_prev->mc_next = cp->mc_next;
    } else {
        mp->mc_first = cp->mc_next;
    }
    pvr2_main_check_removal(mp);
    up(&mp->sem);
}


int pvr2_channel_claim_stream(struct pvr2_channel *cp,
                              struct pvr2_main_stream *sp)
{
    struct pvr2_main *mp = cp->mc_head;
    int code = 0;
    down(&mp->sem);

    if (sp == cp->stream) goto done;
    if (sp->user) {
        code = -EBUSY;
        goto done;
    }
    pvr2_channel_disclaim_stream(cp);
    if (!sp) goto done;
    sp->user = cp;
    cp->stream = sp;
 done:
    up(&mp->sem);
    return code;
}



/*
  Stuff for Emacs to see, in order to encourage consistent editing style:
  *** Local Variables: ***
  *** mode: c ***
  *** fill-column: 75 ***
  *** c-basic-offset: 4 ***
  *** indent-tabs-mode: nil ***
  *** End: ***

  Similarly for vim to see:
  vim:expandtab:textwidth=75

*/
