/*
 *
 *  Copyright (C) 2004 Aurelien Alleaume <slts@free.fr>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/smp_lock.h>
#include <linux/completion.h>
#include <asm/uaccess.h>
#include <linux/usb.h>

#include <linux/slab.h>
#include <linux/usb.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/smp_lock.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/signal.h>
#include <linux/poll.h>
#include <linux/proc_fs.h>

#include <linux/videodev.h>

#include "pvrusb2.h"
#include "pvrusb2-hdw.h"
#include "pvrusb2-buffer.h"
#include "pvrusb2-stream.h"
#include "pvrusb2-debug.h"

#define DRIVER_VERSION "0.7-mci (21-Apr-2005)"
#define DRIVER_AUTHOR "Mike Isely <isely@pobox.com>"
#define DRIVER_DESC "Hauppauge WinTV-PVR-USB2 MPEG2 Encoder/Tuner"

static int video_nr = -1;       // Number of Video Devices Initialized
static int pvr_nr = 0;          // sequential number of pvrusb2 device

module_param(video_nr, int, -1);
MODULE_PARM_DESC(video_nr, "Offset for device's minor");


#ifdef notdef
int debug = (PVR2_TRACE_OPEN_CLOSE|
	     PVR2_TRACE_ERROR_LEGS|
	     PVR2_TRACE_FIRMWARE|
	     PVR2_TRACE_EEPROM |
	     PVR2_TRACE_INIT |
	     PVR2_TRACE_TUNER|
	     PVR2_TRACE_I2C |
	     PVR2_TRACE_STATE |
	     PVR2_TRACE_START_STOP |
	     PVR2_TRACE_AUDIO |
	     PVR2_TRACE_DECODER |
	     PVR2_TRACE_CTL |
	     PVR2_TRACE_ENCODER);
#else
int debug = (PVR2_TRACE_ERROR_LEGS|
	     PVR2_TRACE_FIRMWARE|
	     PVR2_TRACE_EEPROM |
	     PVR2_TRACE_INIT |
	     PVR2_TRACE_I2C |
	     PVR2_TRACE_STATE |
	     PVR2_TRACE_START_STOP |
	     PVR2_TRACE_CTL);
#endif

MODULE_PARM_DESC(debug, "Debug trace mask");

static struct usb_device_id pvr_table[] = {
	{ USB_DEVICE(0x2040, 0x2900) },
	{ }
};

static int pvr_probe(struct usb_interface *intf, const struct usb_device_id *devid)
{
    struct pvr *pvr;
    int i, ret = 0;

    pvr = kmalloc(sizeof(struct pvr), GFP_KERNEL);
    if (!pvr) {
	err("%s: couldn't allocate pvr struct", __FUNCTION__);
	return -ENOMEM;
    }

    pvr2_trace(PVR2_TRACE_INIT,"pvr_probe(pvr=0x%x)",(unsigned int)pvr);

    memset(pvr, 0, sizeof(struct pvr));

    init_MUTEX(&pvr->sem);
    down(&pvr->sem);

    usb_set_intfdata(intf, pvr);

    pvr->hdw = pvr2_hdw_create(intf);
    if (!pvr->hdw) {
	up(&pvr->sem);
	err("Failed to create hdw handler");
	return -ENOMEM;
    }

    if (!pvr2_hdw_init_ok(pvr->hdw)) {
	info("Initialization not complete");
	up(&pvr->sem);
	return 0;
    }


    pvr->nr = pvr_nr++;

    /* register streams */

    pvr->stream[0] = pvr_stream_alloc(pvr, PVR_VID_ENDPOINT,
                                      VFL_TYPE_GRABBER,video_nr,
				      pvr2_config_mpeg);
/*
    pvr->stream[1] = pvr_stream_alloc(pvr, PVR_VBI_ENDPOINT,
                                      VFL_TYPE_VBI, video_nr,
				      pvr2_config_vbi);
    pvr->stream[2] = pvr_stream_alloc(pvr, PVR_VID_ENDPOINT,
                                      VFL_TYPE_RADIO, video_nr,
				      pvr2_config_radio);
*/
    pvr->stream_count = 1;

    for (i = 0; i < pvr->stream_count ; i++) {
        if (pvr->stream[i] == NULL) {
            ret = -ENOMEM;
            break;
        }
    }

    if (ret) {
        /* One stream failed to alloc. Free'em all */
        for (i = 0 ; i < pvr->stream_count ; i++)
            if (pvr->stream[i])
                pvr_stream_free(pvr->stream[i]);
        goto err1;
    }

    up(&pvr->sem);

    return 0;

err1:
    pvr2_hdw_destroy(pvr->hdw);
    up(&pvr->sem);
    kfree(pvr);

    return ret;
}

/*
 * pvr_disconnect()
 *
 */

static void pvr_disconnect(struct usb_interface *intf)
{
    struct pvr *pvr = usb_get_intfdata(intf);
    int i;
    int firmware_flag = !pvr2_hdw_init_ok(pvr->hdw);

    usb_set_intfdata (intf, NULL);

    pvr2_trace(PVR2_TRACE_INIT,"pvr_disconnect(pvr=0x%x)",(unsigned int)pvr);

    pvr2_hdw_disconnect(pvr->hdw);

    if (firmware_flag) {
        /* we did not initialize everything because device was
         * going to re-enumerate
         */
	pvr2_hdw_destroy(pvr->hdw);
        kfree(pvr);
        return;
    }

    down(&pvr->sem);

    for (i = 0 ; i < pvr->stream_count ; i++)
        pvr_stream_free(pvr->stream[i]);

    pvr2_hdw_destroy(pvr->hdw);
    up(&pvr->sem);
    kfree(pvr);
}

static struct usb_driver pvr_driver = {
    owner:   THIS_MODULE,
    name:           "pvrusb2",
    id_table:       pvr_table,
    probe:          pvr_probe,
    disconnect:     pvr_disconnect
};

/*
 * pvr_init() / pvr_exit()
 *
 * This code is run to initialize/exit the driver.
 *
 */

static int __init pvr_init(void)
{
    int ret;

    pvr_debug("pvr_init");

    /* Auto-load tuner and audio modules */
    request_module("tuner");
    request_module("msp3400");
    request_module("saa7115");

    ret = usb_register(&pvr_driver);

    if (ret == 0)
	info(DRIVER_DESC " : " DRIVER_VERSION);
    if (debug) info("Debug mask is %d (0x%x)",debug,debug);

    return ret;
}

static void __exit pvr_exit(void)
{

    pvr_debug("pvr_exit");

    usb_deregister(&pvr_driver);
}

module_init(pvr_init);
module_exit(pvr_exit);

MODULE_DEVICE_TABLE (usb, pvr_table);
MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("GPL");
