/*
 *
 *  Copyright (C) 2005 Mike Isely <isely@pobox.com>
 *  Copyright (C) 2004 Aurelien Alleaume <slts@free.fr>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#ifndef __PVRUSB2_H
#define __PVRUSB2_H


#include <linux/videodev.h>
#include <linux/usb.h>
#include "pvrusb2-hdw.h"

/* Buffer management constants */
#define PVR_VID_BUFFSIZE        PAGE_ALIGN(0x4000)
#define PVR_NBUFF               32       /* Max number of video buffers */

#define PVR_VID_ENDPOINT        0x84
#define PVR_UNK_ENDPOINT        0x86    /* maybe raw yuv ? */
#define PVR_VBI_ENDPOINT        0x88

#define PVR_DEBUG 1
#define pvr_debug(...)    do { if(PVR_DEBUG) info(__VA_ARGS__); } while(0)
/* BUG_ON is a bit aggressive... maybe find a replacement */
#define PVR_BUG_ON(cond)   BUG_ON(cond)

#define VID_HARDWARE_PVRUSB2    38  /* FIXME : need a good value */

#define PVR_NSTREAM 4

enum pvr_memory_state {
    _freed,
    _allocated,
    _mapped,    /* at least one buffer is mapped */
};

enum pvr_io_method {
    _none,
    _read,
    _mmap,
    _uptr
};

struct pvr_mapping {
	struct pvr_stream *stream;
	int count;
};

struct pvr_buffer {
    struct pvr_stream *stream;
    void *buff;     /* Allocated buffer. Page aligned one is in purb */
    u8 *data;     /* First unread data byte */
    size_t size;    /* Size of unread data (for read-io) */
    struct urb *purb;
    short submitted;
    struct pvr_mapping *map;    /* NULL if not mapped */
    int index;
};

struct pvr_uptr {
    unsigned long ptr;
    unsigned int length;
};

struct pvr_stream {
    struct pvr *pvr;
    struct video_device *vdev;
    int v4l_type;

    struct semaphore sem;
    int user;

    unsigned long last_buffer_timestamp;
    enum pvr_memory_state mem_state;
    enum pvr_io_method io_method;
    struct pvr_fh *io_owner;
    enum pvr2_config config;

    int endpoint;   /* usb endpoint to read */
    atomic_t submitted_urbs;
    wait_queue_head_t urbs_collected;
    wait_queue_head_t wait_data;

    struct pvr_buffer buffer[PVR_NBUFF];        /* allocated buffers */
    int buffer_count;   /* how many buffer do we want ? */
    int map_count;      /* number of mapped buffers */
    struct pvr_buffer *rx_buffer[PVR_NBUFF];    /* ring of recv buffers */
    spinlock_t ring_lock;   /* we don't want simultaneous add to the ring. */
    unsigned int rx1;
    /* might want to use atomic for smp ?*/
    unsigned int volatile rx2;
    /* ring for user pointers (acces under sem lock)*/
    struct pvr_uptr userptr[PVR_NBUFF];
    unsigned int uptr1;
    unsigned int uptr2;
    int align;
};

struct pvr_fh {
    struct pvr_stream *stream;
	enum v4l2_priority prio;
};

/* The complete device state */
struct pvr {
    struct pvr2_hdw *hdw;
    struct semaphore sem;
    int nr;
    int supported_std;
    struct v4l2_prio_state prio;

    /* streams */
    struct pvr_stream* stream[PVR_NSTREAM];
    int stream_count;

};

#endif /* __PVRUSB2_H */
